/* @(#) pSOSystem ARM/V2.2.0: bin/source/make2bat/macro.c 2.7 95/12/13 19:32:45 */
/***********************************************************************/
/*                                                                     */
/*   MODULE:  bin/source/make/macro.c                                  */
/*   DATE:    95/12/13                                                 */
/*   PURPOSE: Macro definition, lookup, and expansion                  */
/*                                                                     */
/*---------------------------------------------------------------------*/
/*                                                                     */
/*           Copyright 1991 - 1995, Integrated Systems, Inc.           */
/*                      ALL RIGHTS RESERVED                            */
/*                                                                     */
/*   Permission is hereby granted to licensees of Integrated Systems,  */
/*   Inc. products to use or abstract this computer program for the    */
/*   sole purpose of implementing a product based on Integrated        */
/*   Systems, Inc. products.   No other rights to reproduce, use,      */
/*   or disseminate this computer program, whether in part or in       */
/*   whole, are granted.                                               */
/*                                                                     */
/*   Integrated Systems, Inc. makes no representation or warranties    */
/*   with respect to the performance of this computer program, and     */
/*   specifically disclaims any responsibility for any damages,        */
/*   special or consequential, connected with the use of this program. */
/*                                                                     */
/*---------------------------------------------------------------------*/
/*                                                                     */
/*                                                                     */
/*                                                                     */
/***********************************************************************/
#include "global.h"
#include <string.h>
#include <stdio.h>

extern int E_opt_set;

/*---------------------------------------------------------------------*/
/* The macro table. An entry for a macro in the table has pointers to  */
/* two character strings. One for the name and the other for the       */
/* value of the macro. The last is an attribute telling where the      */
/* was defined: Makefile, Command Line or Environment.                 */
/*---------------------------------------------------------------------*/
struct macro
    {
    struct macro *next;      /* ptr to next macro in list */
    char *name;              /* ptr to name of macro */
    char *value;             /* ptr to value of macro */
    int  attrib;             /* Where defined? M_FILE, M_CMD_LINE, M_ENV */
    };

static struct macro *MacrosList;

/***********************************************************************/
/*   MacroInit: Initialize the macro table                             */
/*                                                                     */
/***********************************************************************/
void MacroInit(void)
{
MacrosList = (struct macro *)0;
}

/***********************************************************************/
/* MacroLookup: Look up a macro, and return a pointer to its value     */
/*                                                                     */
/*      INPUTS: name: pointer to name of macro to look up              */
/*     OUTPUTS: *MacPtr: pointer to macro structure                    */
/*                                                                     */
/*     RETURNS: ptr to expanded value of macro or NULL                 */
/*                                                                     */
/***********************************************************************/
static char *MacroLookup(const char *name, struct macro **MacPtr)
{
struct macro *p;

p = MacrosList;
while (p != (struct macro *)0)
    {
    if (strcmp(name, p->name) == 0)
        {
        *MacPtr = p;
        return(MacroExpand(p->value, strlen(p->value)));
        }
    else
        p = p->next;
    }
return (char *)0;
} 

/***********************************************************************/
/* MacroDefine: Define a macro                                         */
/*                                                                     */
/*      INPUTS: name: ptr to name of macro                             */
/*              value: ptr to value of macro                           */
/*        NOTE: If macro to be defined already exists, old entry may   */
/*              be re-defined!                                         */
/*                                                                     */
/***********************************************************************/
void MacroDefine(const char *name, const char *value, int where)
{
int i;
struct macro *MacroPtr, *NewPtr;
char *ExistingValue;

/*---------------------------------------------------------------------*/
/* If macro is already defined, see whether to the old definition      */
/* needs to be replaced.  If yes, then remove it else return.          */
/*---------------------------------------------------------------------*/
ExistingValue = MacroLookup(name, &MacroPtr);
if (ExistingValue != (char *)0)
    {
    MemFree(ExistingValue);
    switch (MacroPtr->attrib)
        {
        /*
         * - Command line definitions overrides every other definition.
         * - A command line definition overrides a previous command line
         *   definition.
         * - There can never be two similar definition in the environment.
         * - If "-e" option is specified then a definition in environment
         *   overrides a definition in Makefile.
         */
        case M_CMD_LINE:
		    if (where == M_CMD_LINE)
			    break;
            else
			    return;
        case M_ENV:
            if ((where == M_CMD_LINE) || ((where == M_FILE) && !E_opt_set))
                break;
            else
                return;
        case M_FILE:
            if (!E_opt_set && (where == M_ENV))
                return;
            break;
        default:
            Err ("Internal Macro definition error!\n");
            break;
        }
    /*-----------------------------------------------------------------*/
    /* Redefine the existing macro by throwing away the old definition */
    /* and putting in a new one                                        */
    /*-----------------------------------------------------------------*/
    MemFree(MacroPtr->value);
    MacroPtr->value = MemNewString(value);
    MacroPtr->attrib = where;
    return;
    }

/*---------------------------------------------------------------------*/
/* Macro was not previously defined so just create a new entry for it. */
/*---------------------------------------------------------------------*/
NewPtr = (struct macro *)MemAlloc(sizeof(struct macro));
NewPtr->name = MemNewString(name);
NewPtr->value = MemNewString(value);
NewPtr->attrib = where;

/*---------------------------------------------------------------------*/
/* Insert the new macro at the beginning of the list.                  */
/*---------------------------------------------------------------------*/
NewPtr->next = MacrosList;
MacrosList = NewPtr;
}


/***********************************************************************/
/*   StoreChar: Put a character into a dynamically-allocated buffer    */
/*                                                                     */
/*       INPUT: ch: Character to store                                 */
/* IN/OUT-PUTS: *BuffPtrPtr: ptr to beginning of buffer                */
/*              *NextPtrPtr: ptr to next available position in buffer  */
/*              *EndPtrPtr:  ptr to LAST available byte in buffer      */
/*        NOTE: This is slightly tricky because if the current buffer  */
/*              is about to overflow we must obtain a new (larger)     */
/*              one, copy the contents of the current one to it, and   */
/*              return the old one to the dynamic memory manager.      */
/*                                                                     */
/***********************************************************************/
static void StoreChar(char **BuffPtrPtr, char **NextPtrPtr, 
  char **EndPtrPtr, char ch)
{
#define INCREMENT_AMOUNT 10  

char *OldBuffPtr, *p;
int OldBuffLen;

/*---------------------------------------------------------------------*/
/* If next available byte is past the last one in the buffer, its time */
/* to get a new one.                                                   */
/*---------------------------------------------------------------------*/
if (*NextPtrPtr > *EndPtrPtr)
    {
    /*-----------------------------------------------------------------*/
    /* Save ptr to and length of "old" buffer.                         */
    /*-----------------------------------------------------------------*/
    OldBuffPtr = *BuffPtrPtr;
    OldBuffLen = ((int)(*EndPtrPtr - *BuffPtrPtr)) + 1;

    /*-----------------------------------------------------------------*/
    /* Obtain the new buffer and initialize the 3 pointers to it       */
    /*-----------------------------------------------------------------*/
    *BuffPtrPtr = MemAlloc(OldBuffLen + INCREMENT_AMOUNT);
    *NextPtrPtr = *BuffPtrPtr;
    *EndPtrPtr = &((*BuffPtrPtr)[OldBuffLen + INCREMENT_AMOUNT - 1]);

    /*-----------------------------------------------------------------*/
    /* Now copy the contents of the "old" buffer to the "new" one.     */
    /*-----------------------------------------------------------------*/
    p = OldBuffPtr;
    while (OldBuffLen--)
        {
        **NextPtrPtr = *p++;
        (*NextPtrPtr)++;
        }

    /*-----------------------------------------------------------------*/
    /* Give the "old" buffer back for recycling.                       */
    /*-----------------------------------------------------------------*/
    MemFree(OldBuffPtr);
    }

/*---------------------------------------------------------------------*/
/* Store the character and increment "next available byte" pointer     */
/*---------------------------------------------------------------------*/
**NextPtrPtr = ch;
(*NextPtrPtr)++;
}


/***********************************************************************/
/* MacroExpand: Expand macros in a string                              */
/*                                                                     */
/*      INPUTS: InputPtr: pointer to string to be expanded             */
/*              InputLen: length of string                             */
/*     RETURNS: ptr to expanded equivlant of InputPtr[0] through       */
/*              InputPtr[len - 1]                                      */
/*        NOTE: The return value points to a dynamically-allocated     */
/*              memory segment, and it is the caller's responsibility  */
/*              to return the segment when done using it.              */
/***********************************************************************/
#define STORE_OUTPUT_CHAR(c) StoreChar(&OutputStartPtr, &OutputPtr, \
  &OutputEndPtr, c)
#define INC_IN_PTR(pt) if (++pt > InputEndPtr) Err("Illegal expression")

char *MacroExpand(char *InputPtr, int InputLen)
{
char *OutputStartPtr;    /* Points to beginning of output buffer */
char *OutputPtr;         /* Points to next avail. byte in output buffer */
char *OutputEndPtr;      /* Points to last usable byte in output buffer */
char *InputEndPtr;       /* Points to last byte in input string */

char OpeningDelim;       /* Opening delimiter in macro name - ( or { */
char ClosingDelim;       /* Expected closing delimiter - ) or } */
int ExpressionLength;    /* Length of expression between delimiters */
int ParenLevel;          /* Delimiter nesting level */

char *MacroNamePtr; 
char *MacroValuePtr;

char mac1[2];
char *ptr;
int i;
struct macro *JunkPtr;

/*---------------------------------------------------------------------*/
/* First thing to do is allocate an output buffer to put the expanded  */
/* string into.  We don't know how long this will be, so let's use     */
/* the length of the input string as a starting point.  Every time a   */
/* character is put into this buffer, it will be checked for overflow, */
/* and if necessary, a new, larger buffer will replace the current     */
/* one.                                                                */
/*---------------------------------------------------------------------*/
OutputStartPtr = MemAlloc(InputLen);
OutputPtr = OutputStartPtr;
OutputEndPtr = &OutputStartPtr[InputLen - 1];

/*---------------------------------------------------------------------*/
/* The string that this routine evaluates isn't necessarily            */
/* null-terminated, so we will set InputEndPtr to point to the last    */
/* byte in the string to be evaluated.                                 */
/*---------------------------------------------------------------------*/
InputEndPtr = &InputPtr[InputLen - 1];

/*---------------------------------------------------------------------*/
/* Now for the main loop.                                              */
/*---------------------------------------------------------------------*/
while (InputPtr <= InputEndPtr)
    {
    /*-----------------------------------------------------------------*/
    /* If character is not a '$', just copy it to the output buffer.   */
    /*-----------------------------------------------------------------*/
    if (*InputPtr != '$') 
        STORE_OUTPUT_CHAR(*InputPtr++);
    else
        /*-------------------------------------------------------------*/
        /* A '$' indicates a macro name follows                        */
        /*-------------------------------------------------------------*/
        {
        INC_IN_PTR(InputPtr);  /* Throw away the $ */

        if ((*InputPtr != '(') && (*InputPtr != '{'))
            {
            /*---------------------------------------------------------*/
            /* Single-character macros are easy!                       */
            /*---------------------------------------------------------*/
            mac1[0] = *InputPtr++;
            mac1[1] = 0;
            MacroValuePtr = MacroLookup(mac1, &JunkPtr);
            if (MacroValuePtr)
                {
                ptr = MacroValuePtr;
                while (*ptr) STORE_OUTPUT_CHAR(*ptr++);
                MemFree(MacroValuePtr);
                }
            }
        else 
            {
            /*---------------------------------------------------------*/
            /* Macro name is delimited by parentheses, e.g. $(MACRO),  */
            /* or braces, e.g. ${MACRO}.  First thing to do is find    */
            /* the closing delimiter.  This is complicated slightly    */
            /* by the fact that there could be nested delimiters.      */
            /*---------------------------------------------------------*/
            OpeningDelim = *InputPtr;
            ClosingDelim = (OpeningDelim == '(') ? ')' : '}';
            INC_IN_PTR(InputPtr);

            ptr = InputPtr;
            ExpressionLength = 0;
            ParenLevel = 1;

            while (ParenLevel)
                {
                if (*ptr == OpeningDelim)
                    ParenLevel++;
                else if (*ptr == ClosingDelim)
                    ParenLevel--;
                else
                    ExpressionLength++;

                if (ParenLevel != 0)
                    INC_IN_PTR(ptr);
                }

            /*---------------------------------------------------------*/
            /* At this point: InputPtr --> first byte of expression    */
            /*                ptr --> closing delimiter of expression  */
            /* The expression may have nested macros, so we first need */
            /* to "expand" the macro name by calling this routine      */
            /* recursively.   Then put InputPtr past the nested        */
            /* expression.                                             */
            /*---------------------------------------------------------*/
            MacroNamePtr = MacroExpand(InputPtr, (int)(ptr - InputPtr));
            InputPtr = ++ptr;

            /*---------------------------------------------------------*/
            /* Now look up the value of the macro and put it in the    */
            /* output buffer.                                          */
            /*---------------------------------------------------------*/
            MacroValuePtr = MacroLookup(MacroNamePtr, &JunkPtr);
            if (MacroValuePtr)
                {
                ptr = MacroValuePtr;
                while (*ptr) STORE_OUTPUT_CHAR(*ptr++);
                MemFree(MacroValuePtr);
                }
            MemFree(MacroNamePtr);
            }
        }
    }

STORE_OUTPUT_CHAR(0);   /* Null-terminate the string */
return OutputStartPtr;
}


/***********************************************************************/
/*   MacroDump: Print the contents of the macro table                  */
/*                                                                     */
/***********************************************************************/
void MacroDump(void)
{
char *t1, *t2;
struct macro *p;
int i;

p = MacrosList;
i = 0;
printf("Macro table:\n");
while (p != (struct macro *)0)
    {
    printf("(%d) \"%s\"=\"%s\" \n", i++, p->name, p->value);
    p = p->next;
    }
}
