/* @(#) pSOSystem ARM/V2.2.0: bin/source/make2bat/rules.c 2.5 95/12/13 19:32:49 */
/***********************************************************************/
/*                                                                     */
/*   MODULE:  bin/source/make/rules.c                                  */
/*   DATE:    95/12/13                                                 */
/*   PURPOSE: Functions related to storage and retrieval of rules      */
/*                                                                     */
/*---------------------------------------------------------------------*/
/*                                                                     */
/*           Copyright 1991 - 1995, Integrated Systems, Inc.           */
/*                      ALL RIGHTS RESERVED                            */
/*                                                                     */
/*   Permission is hereby granted to licensees of Integrated Systems,  */
/*   Inc. products to use or abstract this computer program for the    */
/*   sole purpose of implementing a product based on Integrated        */
/*   Systems, Inc. products.   No other rights to reproduce, use,      */
/*   or disseminate this computer program, whether in part or in       */
/*   whole, are granted.                                               */
/*                                                                     */
/*   Integrated Systems, Inc. makes no representation or warranties    */
/*   with respect to the performance of this computer program, and     */
/*   specifically disclaims any responsibility for any damages,        */
/*   special or consequential, connected with the use of this program. */
/*                                                                     */
/*---------------------------------------------------------------------*/
/*                                                                     */
/*                                                                     */
/*                                                                     */
/***********************************************************************/
#include "global.h"
#include <stdio.h>
#include <string.h>

struct dependency
    {
    struct dependency *next;
    char *name;
    };

struct action
    {
    struct action *next;
    char *name;
    };

struct rule
    {
    struct rule *next;            /* ptr to next rule or NULL */
    char built;
    char *target;                 /* pointer to target name */
    struct dependency *deplist;
    struct action *actlist;
    };

static struct rule *RulesList;

/***********************************************************************/
/*   RulesInit: Initialize this module                                 */
/*                                                                     */
/***********************************************************************/
void RulesInit(void)
{
RulesList = (struct rule *)0;
}


/***********************************************************************/
/*    findrule: Find an existing rule and return a pointer to it       */
/*                                                                     */
/*      INPUTS: name: ptr to name of target                            */
/*                                                                     */
/*     RETURNS: ptr to rule structure                                  */
/*        NOTE: Rule is expected to be defined.  Fatal error is        */
/*              generated if it is not                                 */
/***********************************************************************/
static struct rule *findrule(char *name)
{
struct rule *p;

p = RulesList;
while (p != (struct rule *)0)
    {
    if (strcmp(p->target, name) == 0)
        return p;
    else
        p = p->next;
    }
ErrStr("Cannot find rule for %s", name);
return (struct rule *)0;
}

/***********************************************************************/
/*   RulesFind:                                                        */
/*                                                                     */
/*      INPUTS: name: ptr to name of target                            */
/*              NotBuiltFlag: If true, only looking for non-built      */
/*                 targets.  In this case FALSE will be returned       */
/*                 in the case where a target's rule is found but has  */
/*                 already been marked as "built".                     */
/*     RETURNS: TRUE if rule found, FALSE if not                       */
/*     NOTE(S):                                                        */
/*                                                                     */
/***********************************************************************/
int RulesFind(char *name, char NotBuiltFlag)
{
struct rule *p;

p = RulesList;
while (p != (struct rule *)0)
    {
    if (strcmp(p->target, name) == 0)
        {
        if (NotBuiltFlag && p->built) return 0;
        else return 1;
        }
    else
        p = p->next;
    }
return 0;
}


/***********************************************************************/
/* RulesMarkBuilt: Mark a target as having been built                  */
/*                                                                     */
/*      INPUTS: target: ptr to name of target                          */
/*                                                                     */
/***********************************************************************/
void RulesMarkBuilt(char *target)
{
findrule(target)->built = 1;
}

/***********************************************************************/
/* RulesFindDependency: Return the name of a dependency for a target   */
/*                                                                     */
/*      INPUTS: target: ptr to name of target                          */
/*              index: dependency # (starting from 1)                  */
/*     RETURNS: ptr to name of dependency, or NULL if index too high   */
/*     NOTE(S): non-NULL return value points to a dynamically allocated*/
/*              segment, and it is the caller's responsibility to      */
/*              return it properly when done                           */
/***********************************************************************/
char *RulesFindDependency(char *target, int index)
{
struct dependency *dp;

dp = findrule(target)->deplist;
if (dp == (struct dependency *)0) return (char *)0;

while (--index)
    {
    if (dp->next == (struct dependency *)0) return (char *)0;
    else dp = dp->next;
    }
return (MemNewString(dp->name));
}

/***********************************************************************/
/* RulesFindAction: Return the name of an action for a target          */
/*                                                                     */
/*      INPUTS: target: ptr to name of target                          */
/*              index: action # (starting from 1)                      */
/*     RETURNS: ptr to name of action, or NULL if index too high       */
/*     NOTE(S): non-NULL return value points to a dynamically allocated*/
/*              segment, and it is the caller's responsibility to      */
/*              return it properly when done                           */
/***********************************************************************/
char *RulesFindAction(char *target, int index)
{
struct action *ap;

ap = findrule(target)->actlist;
if (ap == (struct action *)0) return (char *)0;

while (--index)
    {
    if (ap->next == (struct action *)0) return (char *)0;
    else ap = ap->next;
    }
return(MemNewString(ap->name));
}


/***********************************************************************/
/*   RulesMake: Make a new entry in the rules table                    */
/*                                                                     */
/*      INPUTS: target: pointer to name of target                      */
/*                                                                     */
/***********************************************************************/
void RulesMake(char *target)
{
struct rule *NewPtr;

NewPtr = (struct rule *)MemAlloc(sizeof(struct rule));
NewPtr->built = 0;
NewPtr->target = MemNewString(target);
NewPtr->deplist = (struct dependency *)0;
NewPtr->actlist = (struct action *)0;

/*---------------------------------------------------------------------*/
/* Insert the new rule at the beginning of the list.                   */
/*---------------------------------------------------------------------*/
NewPtr->next = RulesList;
RulesList = NewPtr;
}

/***********************************************************************/
/* RulesAddDependency: Add a dependency to an existing rule            */
/*                                                                     */
/*      INPUTS: target: ptr to name of target                          */
/*              dependency: ptr to name of dependency                  */
/*                                                                     */
/***********************************************************************/
void RulesAddDependency(char *target, char *dependency)
{
int i;
struct dependency *NewDepPtr, *DepPtr;
struct rule *RulePtr;

RulePtr = findrule(target);

/*-------------------------------------------------------------------*/
/* Allocate a new dependency structure and fill it in.               */
/*-------------------------------------------------------------------*/
NewDepPtr = (struct dependency *)MemAlloc(sizeof(struct dependency));
NewDepPtr->name = MemNewString(dependency);
NewDepPtr->next = (struct dependency *)0;

/*-------------------------------------------------------------------*/
/* Put the new dependency at the end of the list.
/*-------------------------------------------------------------------*/
if (RulePtr->deplist == (struct dependency *)0)
    RulePtr->deplist = NewDepPtr;
else
    {
    DepPtr = RulePtr->deplist;
    while (DepPtr->next != (struct dependency *)0)
        DepPtr = DepPtr->next;
    DepPtr->next = NewDepPtr;
    }
}

/***********************************************************************/
/* RulesAddAction: Add an action to an existing rule                   */
/*                                                                     */
/*      INPUTS: target: ptr to name of target                          */
/*              action: ptr to action string                           */
/*                                                                     */
/***********************************************************************/
void RulesAddAction(char *target, char *action)
{
int i;
struct action *NewActPtr, *ActPtr;
struct rule *RulePtr;

RulePtr = findrule(target);

/*-------------------------------------------------------------------*/
/* Allocate a new action structure and fill it in.               */
/*-------------------------------------------------------------------*/
NewActPtr = (struct action *)MemAlloc(sizeof(struct action));
NewActPtr->name = MemNewString(action);
NewActPtr->next = (struct action *)0;

/*-------------------------------------------------------------------*/
/* Put the new action at the end of the list.
/*-------------------------------------------------------------------*/
if (RulePtr->actlist == (struct action *)0)
    RulePtr->actlist = NewActPtr;
else
    {
    ActPtr = RulePtr->actlist;
    while (ActPtr->next != (struct action *)0)
        ActPtr = ActPtr->next;
    ActPtr->next = NewActPtr;
    }
}

/***********************************************************************/
/*   RulesDump: Dump the rules table                                   */
/*                                                                     */
/***********************************************************************/
void RulesDump(void)
{
int i;
struct rule *p;
struct dependency *dp;
struct action *ap;

printf("Rules Table:\n");

p = RulesList;
while (p != (struct rule *)0) 
    {
    printf("RULE FOR TARGET: %s\n", p->target);
    printf("  Dependencies:\n");
    dp = p->deplist;
    while (dp != (struct dependency *)0)
        {
        printf("     %s\n", dp->name);
        dp = dp->next;
        }
    printf("  Actions:\n");
    ap = p->actlist;
    while (ap != (struct action *)0)
        {
        printf("     %s\n", ap->name);
        ap = ap->next;
        }
    p = p->next;
    }
}
