/* @(#) pSOSystem ARM/V2.2.1: bin/source/smove.c  1.2 97/05/14 15:51:55 */
/***********************************************************************/
/*                                                                     */
/*   MODULE:  bin/source/smove.c                                       */
/*   DATE:    97/05/14                                                 */
/*   PURPOSE: S-record relocation utility                              */
/*                                                                     */
/*---------------------------------------------------------------------*/
/*                                                                     */
/*         Copyright 1991 -  1994, Integrated Systems, Inc.            */
/*                      ALL RIGHTS RESERVED                            */
/*                                                                     */
/*   Permission is hereby granted to licensees of Integrated Systems,  */
/*   Inc. products to use or abstract this computer program for the    */
/*   sole purpose of implementing a product based on Integrated        */
/*   Systems, Inc. products.   No other rights to reproduce, use,      */
/*   or disseminate this computer program, whether in part or in       */
/*   whole, are granted.                                               */
/*                                                                     */
/*   Integrated Systems, Inc. makes no representation or warranties    */
/*   with respect to the performance of this computer program, and     */
/*   specifically disclaims any responsibility for any damages,        */
/*   special or consequential, connected with the use of this program. */
/*                                                                     */
/*---------------------------------------------------------------------*/
/*                                                                     */
/*  This program is a utility which will change the load address of    */
/*  Motorola S-record files.  It is the successor to "soff".           */
/*                                                                     */
/*  This module is designed to be as environment-independent as        */
/*  possible.  All actual input and output is done in a few small      */
/*  routines which can be easily modified to interface with the        */
/*  execution environment.  The only routines which may need           */
/*  modification are charout() and getnext().  The requirements for    */
/*  each of the routines are explained in their headers.               */
/*                                                                     */
/*  As supplied, this module should work correctly on most UNIX        */
/*  systems.  It is assumed to use standard input and output for the   */
/*  input and output files, respectively.                              */
/*                                                                     */
/*  The syntax is:                                                     */
/*     smove <address_range> <dest_addr> [s9] <infile> <outfile>       */
/*  where <address_range> specifies a range of S-record addresses      */
/*  which will be relocated to begin at <dest_addr>.  <address_range>  */
/*  consists of two hex addresses separated by two periods, e.g.       */
/*  f000..ffff                                                         */
/*                                                                     */
/*  Assuming "file.srec" is the s-record input file, and "file.new"    */
/*  is the altered S-record file to be produced, command line on a     */
/*  UNIX system to relocate all S-records in the range 0x2000 through  */
/*  0x8fff to 0x100000 would be:                                       */
/*                                                                     */
/*    smove 2000..8fff 100000 file.srec file.new                       */
/*                                                                     */
/*  The s9 flag tells the program to put an S9 record at the end of    */
/*  its output.                                                        */
/*                                                                     */
/*  The only other modification which may be necessary in this module  */
/*  involves the data types.  UINT32 must be defined to be a 32-bit    */
/*  quantities.                                                        */
/*                                                                     */
/*                                                                     */
/*  IMPORTANT NOTE: This utility does not work correctly for S-records */
/*  that are partially in the range to be moved.  This is something    */
/*  that should be fixed in the future.                                */
/*                                                                     */
/***********************************************************************/
#include <stdio.h>

/*---------------------------------------------------------------------*/
/* Data types - may need to be modified on some systems.               */
/*---------------------------------------------------------------------*/
#define UINT32      unsigned long
#define UCHAR       unsigned char

/*---------------------------------------------------------------------*/
/* Constants                                                           */
/*---------------------------------------------------------------------*/
#define TRUE      1
#define FALSE     0
#define CHKSUM    2

static FILE *Src_File, *Dst_File;
/*---------------------------------------------------------------------*/
/* Template for an s-record structure.  The data from each s-record    */
/* is stored in a structure of this type for processing.               */
/*---------------------------------------------------------------------*/
struct s_rec
    {
    UINT32 length;          /* # of loadable bytes */
    UINT32 addr;            /* starting address */
    UCHAR data[512];        /* actual data in ASCII hex */
    };                      /* high nibble, then low nibble */

static UCHAR cs;            /* computed checksum */

/***********************************************************************/
/*     charout: Send a character to the output stream.                 */
/*                                                                     */
/*      INPUTS: c: character to send                                   */
/*                                                                     */
/***********************************************************************/
static void charout(c)
UCHAR c;
{
putchar(c);
}

/***********************************************************************/
/*     getnext:  Read one character from the input stream              */
/*                                                                     */
/*     RETURNS: Character read, or EOF if end-of-file                  */
/*                                                                     */
/***********************************************************************/
static UCHAR getnext()
{
return ((UCHAR) fgetc(Src_File));
}

/***********************************************************************/
/*      msgout: Send a message to the output file.                     */
/*                                                                     */
/*      INPUTS: str_ptr: points to beginning of null-terminated string */
/***********************************************************************/
static void msgout(str_ptr)
UCHAR *str_ptr;
{
UCHAR c;
while (c = *str_ptr++)
    charout(c);
}

/***********************************************************************/
/* ascii_to_hex: Convert ASCII hex character to binary                 */
/*                                                                     */
/*      INPUTS: byte: ASCII character to convert                       */
/*     RETURNS: binary value of hex representation                     */
/***********************************************************************/
static UINT32 ascii_to_hex(byte)
UCHAR byte;
{
if ((byte >= '0') && (byte <= '9'))
    return byte - '0';
else if ((byte >= 'A') && (byte <= 'F'))
    return byte - 'A' + 10;
else
    return -1;
}

/***********************************************************************/
/*    nextpair: Compute the binary value of the next character pair    */
/*              from the input file.                                   */
/*                                                                     */
/*     RETURNS: Binary value of the character pair                     */
/*        NOTE: Consumes 2 bytes from the input stream                 */
/***********************************************************************/
static UINT32 nextpair()
{
UINT32 tmp;
tmp  = ascii_to_hex(getnext()) * 16;
tmp += ascii_to_hex(getnext());
return tmp;
}

/***********************************************************************/
/*      find_s: Find next type 1, 2, or 3 s-record in the input file   */
/*                                                                     */
/*     RETURNS: Type of record found (1, 2, or 3) or 0 if no more.     */
/*        NOTE: Input file is left positioned after the "record type"  */
/*              field in the record which was found.                   */
/***********************************************************************/
static UINT32 find_s()
{
UINT32 type;
char c;

type = 0;
while ((type < 1) || (type > 3))
    {
    c = getnext();
    if (c == 'S')
        type = ascii_to_hex(getnext());
    else
       if (feof(Src_File))
          return 0;
    }

return type;
}

/***********************************************************************/
/*      read_s:  Read next type 1, 2, or 3 s-record from the input     */
/*                                                                     */
/*     RETURNS: TRUE if record is found, FALSE if there are no more    */
/*     OUTPUTS: *s contains information from the s-record              */
/***********************************************************************/
static UCHAR read_s(s)
struct s_rec *s;
{
UINT32 chksum;
UINT32 type, n;
UCHAR *ptr;
UINT32 tmpvar, pairval;

if (type = find_s())  /* If there's another s-record.... */
    {
    /*-----------------------------------------------------------------*/
    /* Get the length of the record and convert it to the number of    */
    /* loadable bytes.                                                 */
    /*-----------------------------------------------------------------*/
    s->length = nextpair();
    chksum = s->length;
    s->length -= (type + 2);

    /*-----------------------------------------------------------------*/
    /* Get the starting address of the record, which could be 2, 3, or */
    /* 4 character pairs depending on record type.                     */
    /*-----------------------------------------------------------------*/
    type += 1;  /* # of pairs is 1 more than "type". */
    s->addr = 0;
    while (type--)
        {
        pairval = nextpair();
        s->addr = s->addr * 256 + pairval;
        chksum += pairval;
        }

    /*-----------------------------------------------------------------*/
    /* s->length contains the number of loadable bytes in the record.  */
    /* Each "loadable byte" is represented by 2 ASCII characters, so   */
    /* read the next s->length * 2 characters from the file into the   */
    /* s->data[] array.                                                */
    /*-----------------------------------------------------------------*/
    n = s->length;
    ptr = &(s->data[0]);
    while (n--)
        {
        *ptr = 16 * ascii_to_hex(getnext());
        *ptr++ += ascii_to_hex(getnext());
        }

    /*-----------------------------------------------------------------*/
    /* Checksum the characters just read.                              */
    /*-----------------------------------------------------------------*/
    ptr = &(s->data[0]);
    for (n = 0; n < s->length; n++)
        chksum += *ptr++;

    /*-----------------------------------------------------------------*/
    /* Finally, finish our computation of what the checksum should be  */
    /* and compare with the value read from the end of the s-record.   */
    /*-----------------------------------------------------------------*/
    chksum = 0xff & (~chksum);
    if (nextpair() != (UINT32)(chksum))
        {
        msgout("Checksum error\n");
        exit(-1);
        }
    return TRUE;
    }

/*---------------------------------------------------------------------*/
/* If there were no more s-records in the file, simply return FALSE.   */
/*---------------------------------------------------------------------*/
else
   return FALSE;
}

/***********************************************************************/
/*     putcsbh: Convert a binary byte value to ascii and write it to   */
/*              the host channel.  Include the characters sent with    */
/*              the current record checksum.                           */
/*                                                                     */
/*      INPUTS: value: Binary value to write.                          */
/*                                                                     */
/***********************************************************************/
static void putcsbh(value)
UCHAR value;
{
fprintf(Dst_File, "%02X", value);
cs += value;
}

/***********************************************************************/
/*     write_s: Write an S-record to the standard output               */
/*                                                                     */
/*      INPUTS: s: pointer to structure specifying the S-record        */
/*                                                                     */
/***********************************************************************/
static void write_s(s)
struct s_rec *s;
{
UINT32 size;
int i;

/*---------------------------------------------------------------------*/
/* Determine the S-record size (type) based on the magnitude of        */
/* its address.                                                        */
/*---------------------------------------------------------------------*/
if (s->addr < 0x10000)
   size = 2;
else
   if (s->addr < 0x1000000)
      size = 3;
   else
      size = 4;

/*---------------------------------------------------------------------*/
/* Print the first two characters of the S-record and initialize       */
/* the record's checksum.                                              */
/*---------------------------------------------------------------------*/
fprintf(Dst_File, "S%X", size - 1);

cs = 0;

/*---------------------------------------------------------------------*/
/* Send record length field.                                           */
/*---------------------------------------------------------------------*/
putcsbh((UCHAR) (s->length + size + 1));

/*---------------------------------------------------------------------*/
/* Send address field.                                                 */
/*---------------------------------------------------------------------*/
if (size == 4) putcsbh((UCHAR)((s->addr >> 24) & 0xFF));
if (size > 2) putcsbh((UCHAR)((s->addr >> 16) & 0xFF));
putcsbh((UCHAR)((s->addr >> 8) & 0xFF));
putcsbh((UCHAR)(s->addr & 0xFF));

/*---------------------------------------------------------------------*/
/* Send data fields.                                                   */
/*---------------------------------------------------------------------*/
for (i = 0; i < s->length; i++)
    putcsbh((UCHAR) s->data[i]);

/*---------------------------------------------------------------------*/
/* Send check-sum and trailing carriage return line feed.              */
/* Increment record count.                                             */
/*---------------------------------------------------------------------*/
putcsbh(~cs);

fprintf(Dst_File, "\n");
}

/***********************************************************************/
/*                                                                     */
/***********************************************************************/
int main(argc, argv)
int argc;
char *argv[];
{
int i, j;
UINT32 SrcBegin, SrcEnd, DstBegin, offset;
UCHAR  s9flag, subflag;
struct s_rec srec;   /* info from current s-record */

/*---------------------------------------------------------------------*/
/* Check number of arguments                                           */
/*---------------------------------------------------------------------*/
if ((argc != 5) && (argc != 6))
   usage();

/*---------------------------------------------------------------------*/
/* Parse command-line arguments                                        */
/*---------------------------------------------------------------------*/
sscanf(argv[1], "%lx..%lx", &SrcBegin, &SrcEnd);
sscanf(argv[2], "%lx",      &DstBegin);

s9flag = 0;
if ((argc == 6) && ((argv[3][0] == 'S') || (argv[3][0] == 's')) && 
                                                            (argv[3][1] == '9'))
   s9flag = 1;

fprintf(stderr,
       "\nMoving records in range 0x%lx through 0x%lx to 0x%lx\n", 
        SrcBegin, SrcEnd, DstBegin);

if (s9flag != 0)
   fprintf(stderr,
          "S9 record will be appended to end of %s\n", argv[(argc - 1)]);

if (SrcBegin > DstBegin)
   {
   subflag = 1;
   offset  = SrcBegin - DstBegin;
   }
else
   {
   if (DstBegin > SrcBegin)
      {
      subflag = 0;
      offset  = DstBegin - SrcBegin;
      }
   else
      {
      subflag = 0;
      offset  = 0;
      }
   }

/*---------------------------------------------------------------------*/
/* Open the files.                                                     */
/*---------------------------------------------------------------------*/
if ((Src_File = fopen(argv[(argc - 2)], "r")) == 0)
   {
   usage();
   return(2);
   }

if ((Dst_File = fopen(argv[(argc - 1)], "w")) == 0)
   {
   usage();
   return(2);
   }

/*---------------------------------------------------------------------*/
/* Now do the work.                                                    */
/*---------------------------------------------------------------------*/
while (read_s(&srec))
      {
      if ((srec.addr >= SrcBegin) && (srec.addr <= SrcEnd))
         {
         if (subflag)
            srec.addr -= offset;
         else
            srec.addr += offset;
         }
    write_s(&srec);
    }

/*---------------------------------------------------------------------*/
/* Hit end of the input.  Append an S9 if it was requested on cmd line */
/*---------------------------------------------------------------------*/
if (s9flag)
   fprintf(Dst_File, "S9030000FC\n");

/*---------------------------------------------------------------------*/
/* Close files.                                                        */
/*---------------------------------------------------------------------*/
fclose(Src_File);
fclose(Dst_File);

/*---------------------------------------------------------------------*/
/* Return Okay Status                                                  */
/*---------------------------------------------------------------------*/
exit(0);
}

/***********************************************************************/
/*       usage: Prints some lines showing the program revision number, */
/*              the reason the command line string was rejected, and   */
/*              the proper command syntax.  Calls exit().              */
/*                                                                     */
/***********************************************************************/
usage()
{
fprintf(stderr, "SMOVE S-record offset utility Version 1.1 (21-AUG-95)\n");
fprintf(stderr,
  "Usage: smove <src_begin>..<src_end> <dst_begin> [s9] <infile> <outfile>\n");
exit(0);
}
