/* @(#) pSOSystem ARM/V2.2.1: bin/source/splitrom.c  1.2 97/05/14 15:51:57 */
/***********************************************************************/
/*                                                                     */
/*   MODULE:  bin/source/splitrom.c                                    */
/*   DATE:    97/05/14                                                 */
/*   PURPOSE: To split an S-record representing the output of a make-  */
/*            file into 2 or 4 S-records representing address zero     */
/*            based ROM images, suitable for programming into EPROMs.  */
/*                                                                     */
/*---------------------------------------------------------------------*/
/*                                                                     */
/*           Copyright 1991, 1992, Integrated Systems, Inc.            */
/*                      ALL RIGHTS RESERVED                            */
/*                                                                     */
/*   Permission is hereby granted to licensees of Integrated Systems,  */
/*   Inc. products to use or abstract this computer program for the    */
/*   sole purpose of implementing a product based on Integrated        */
/*   Systems, Inc. products.   No other rights to reproduce, use,      */
/*   or disseminate this computer program, whether in part or in       */
/*   whole, are granted.                                               */
/*                                                                     */
/*   Integrated Systems, Inc. makes no representation or warranties    */
/*   with respect to the performance of this computer program, and     */
/*   specifically disclaims any responsibility for any damages,        */
/*   special or consequential, connected with the use of this program. */
/*                                                                     */
/*---------------------------------------------------------------------*/
/* Usage:                                                              */
/*                                                                     */
/*  splitrom <in> <b|w|ws> <base> <size> <out1> [<out2> [<out3> <out4>]] */
/*                                                                     */
/* where: <in> is the name of an S-record file containing some abso-   */
/*        lute code you want to program into EPROMs.                   */
/*                                                                     */
/*        <b|w|ws> specifies whether the target EPROMs are byte wide   */
/*        or word wide (16-bits).  "ws" indicates EPROMs that are word */
/*        wide, but that will be used in a system that swaps the order */
/*        in which the bytes are read from EPROM.  i.e., the lower     */
/*        byte is read from EPROM data lines 8-15 and the higher byte  */
/*        is read from EPROM data lines 0-7.  This is how EPROMs for   */
/*        the MVME167 board, for example, are used.                    */
/*                                                                     */
/*        <base> is the base address at which the EPROMs will appear   */
/*        in the target board's memory map.  This is an eight charac-  */
/*        ter hexidecimal address without any prefixes (neither "0x"   */
/*        nor "$").                                                    */
/*                                                                     */
/*        <size> is a decimal number representing the size, in kilo-   */
/*        bytes, of the target EPROMs.                                 */
/*                                                                     */
/*        <out1> ... <out4> are the names of output files.  These will */
/*        contain S-record EPROM images at a base address of zero.     */
/*        The number of files present determines the number of ways    */
/*        that the input file is split.  The number of output files    */
/*        must be one, two, or four.  It should be the number of       */
/*        EPROMs that your target system can access simultaneously.    */
/*                                                                     */
/*        If there are two output files, <out1> will contain the lower */
/*        bytes (or words for 16-bit EPROMs) and <out2> will contain   */
/*        the upper bytes (or words).  If there are four output files, */
/*        the lower lower bytes will be in <out4>, the lower middle    */
/*        bytes will be in <out3>, the upper middle bytes in <out3>,   */
/*        and the upper upper bytes in <out1>.                         */
/*                                                                     */
/*        Note: The specification of both word-wide EPROMs and four    */
/*        output files is not permitted.  None of the supported boards */
/*        can simultaneously access 64-bits (16 x 4).                  */
/*                                                                     */
/*                                                                     */
/* Below are example command lines for several supported boards:       */
/*                                                                     */
/* MVME147: (for 27512 EPROMs)                                         */
/*        splitrom app.hex b FF800000 128 app.u1 app.u2                */
/*                                                                     */
/* CPU-30: (for 27512 EPROMs)                                          */
/*        splitrom app.hex b FF000000 64 app.j4 app.j6 app.j8 app.j10  */
/*                                                                     */
/* MVME167: (for 27C202 EPROMs)                                        */
/*        splitrom app.hex ws FF800000 128 app.u6 app.u8               */
/*                                                                     */
/* CPU-40: (for 27210 EPROMs)                                          */
/*        splitrom app.hex w FF000000 128 app.j29 app.j30              */
/*                                                                     */
/***********************************************************************/
#include <stdio.h>

/*---------------------------------------------------------------------*/
/* Symbol/Macro Definitions                                            */
/*---------------------------------------------------------------------*/
#define OK          0
#define NOT_OK      1
#define ARRAY_SIZE  512     /* Must be kept even for swap() */
#define ULONG       unsigned long
#define even(x)     (((x) & 1) == 0)
#define odd(x)      (((x) & 1) == 1)
#define in_range(x) (((x) >= Base_Address) && \
                     ((x) <= Base_Address + EPROM_Size * Num_Splits - 1))

/*---------------------------------------------------------------------*/
/* Structure Definitions                                               */
/*---------------------------------------------------------------------*/
typedef struct s_record
    {
    char  type;                 /* 1, 2, or 3 */
    ULONG length;               /* # of loadable bytes */
    ULONG address;              /* starting address */
    char  data[ARRAY_SIZE];     /* actual data in ASCII hex */
    } S_RECORD;

/*---------------------------------------------------------------------*/
/* Global Variables                                                    */
/*---------------------------------------------------------------------*/
static int Num_Splits;
static ULONG Base_Address;
static ULONG EPROM_Size;
static enum {BYTE, WORD, WORD_SWAP} Width;
static FILE *Src_File, *Out_File[4], *Output_File;
static char *File_Name1, *File_Name2, *File_Name3, *File_Name4;
static char *Usage_String;
static S_RECORD S_record[5];    /* 1 for input, 4 for output */
static int Sum, Check_Sum;
static char Encountered_out_of_range_data = 0;
static ULONG Byte_Count;

/***********************************************************************/
/* read_params: Read the command line parameters, and set global flags */
/*              accordingly.                                           */
/*                                                                     */
/*      INPUTS: argc: count of arguments present on command line       */
/*              argv: array of pointers to command line strings        */
/*                                                                     */
/*     RETURNS: OK, if everything is read alright; a non-zero error    */
/*              code otherwise.                                        */
/***********************************************************************/
static int read_params(argc, argv)
int argc;
char *argv[];
{
ULONG size;

/*---------------------------------------------------------------------*/
/* Check that the number of parameters is 5, 6 or 8.                   */
/*---------------------------------------------------------------------*/
if ((argc != 6) && (argc != 7) && (argc != 9))
    return 1;

/*---------------------------------------------------------------------*/
/* Check that we can open the input file.                              */
/*---------------------------------------------------------------------*/
Src_File = fopen(argv[1], "r");
if (Src_File == 0)
    {
    Usage_String = argv[1];
    return 2;
    }

/*---------------------------------------------------------------------*/
/* Check that the width parameter is "b", "w", or "ws".                */
/*---------------------------------------------------------------------*/
if (strcmp(argv[2], "b") == 0)
    Width = BYTE;
else if (strcmp(argv[2], "w") == 0)
    Width = WORD;
else if (strcmp(argv[2], "ws") == 0)
    Width = WORD_SWAP;
else
    return 3;

/*---------------------------------------------------------------------*/
/* Check that we can read the base address and EPROM size.             */
/*---------------------------------------------------------------------*/
if (sscanf(argv[3], "%lx", &Base_Address) == EOF)
    return 4;

if (sscanf(argv[4], "%ld", &size) == EOF)
    return 5;
EPROM_Size = size * 1024;

/*---------------------------------------------------------------------*/
/* Check that we can open the output files.                            */
/*---------------------------------------------------------------------*/
Out_File[0] = fopen(argv[5], "w");
if (Out_File[0] == 0)
    {
    Usage_String = argv[5];
    return 6;
    }

if (argc > 6)
    {
    Out_File[1] = fopen(argv[6], "w");
    if (Out_File[1] == 0)
        {
        Usage_String = argv[6];
        return 6;
        }
    }

if (argc == 9)
    {
    Out_File[2] = fopen(argv[7], "w");
    if (Out_File[2] == 0)
        {
        Usage_String = argv[7];
        return 6;
        }

    Out_File[3] = fopen(argv[8], "w");
    if (Out_File[3] == 0)
        {
        Usage_String = argv[8];
        return 6;
        }
    }

/*---------------------------------------------------------------------*/
/* Assign global pointers to the file names for remove_files().        */
/*---------------------------------------------------------------------*/
File_Name1 = argv[5];
File_Name2 = argv[6];
File_Name3 = argv[7];
File_Name4 = argv[8];

/*---------------------------------------------------------------------*/
/* Set Num_Splits based on the number of arguments.                    */
/*---------------------------------------------------------------------*/
Num_Splits = argc - 5;   /* Num_Splits is 1, 2, or 4 */

return OK;
}

/***********************************************************************/
/* remove_files: Removes all the output files.  Called if fatal error  */
/*               occurs.                                               */
/*                                                                     */
/***********************************************************************/
static void remove_files()
{
#ifndef SUN3
remove(File_Name1);

if (Num_Splits > 1)
    remove(File_Name2);

if (Num_Splits == 4)
    {
    remove(File_Name3);
    remove(File_Name4);
    }
#endif /* SUN3 */
}

/***********************************************************************/
/*  show_usage: Prints some lines showing the program revision number, */
/*              the reason the command line string was rejected, and   */
/*              the proper command syntax.                             */
/*                                                                     */
/*      INPUTS: problem_num: an integer encoding the reason that the   */
/*              command line was rejected.                             */
/***********************************************************************/
static void show_usage(problem_num)
int problem_num;
{
printf("SPLITROM S-record to ROM image utility Version 1.1 (01-MAY-93)\n");
switch(problem_num)
    {
    case 1:
        printf("Wrong number of parameters; must be 5, 6, or 8.\n");
        break;
    case 2:
        printf("Unable to open input file \"%s\".\n", Usage_String);
        break;
    case 3:
        printf("The EPROM width must be either \"b\", \"w\", or \"ws\".\n");
        break;
    case 4:
        printf("The EPROM base address must be a hexidecimal number.\n");
        break;
    case 5:
        printf("The EPROM kilobyte size must be a decimal number.\n");
        break;
    case 6:
        printf("Unable to create output file \"%s\".\n", Usage_String);
        break;
    default: break;
    }
printf("Usage: splitrom <in> <b|w|ws> <base> <size> <out1> [<out2>");
printf(" [<out3> <out4>]]\n");
}

/***********************************************************************/
/*  hex_to_int: Convert ASCII hex character to binary.                 */
/*                                                                     */
/*      INPUTS: c: ASCII hex character to convert.                     */
/*     RETURNS: Integer value of the ASCII hex character.              */
/*        NOTE: Fatal error results if input is not a hex character.   */
/***********************************************************************/
static int hex_to_int(c)
int c;
{
if ((c >= '0') && (c <= '9'))
    return c - '0';
if ((c >= 'A') && (c <= 'F'))
    return c - 'A' + 10;
else
    {
    printf("Fatal Error: Found non-hex character inside S-record!\n");
    remove_files();
    exit(-1);
    }
}

/***********************************************************************/
/*    nextbyte: Convert the next 2 hex ASCII characters to an integer. */
/*                                                                     */
/*     RETURNS: Integer value of the character pair.                   */
/*        NOTE: Consumes 2 characters from the input file.             */
/***********************************************************************/
static int nextbyte()
{
int packed_byte;

packed_byte = hex_to_int(fgetc(Src_File));
packed_byte = (packed_byte << 4) + hex_to_int(fgetc(Src_File));
Sum += packed_byte;
return packed_byte;
}

/***********************************************************************/
/*   find_S123: Find next type 1, 2, or 3 S-record in the input file.  */
/*                                                                     */
/*     RETURNS: 1, 2, or 3 if an S-record is found, 0 if not found.    */
/*        NOTE: file is left positioned after the "record type" field. */
/***********************************************************************/
static int find_S123()
{
int c, type = 0;

while((type < 1) || (type > 3))
    {
    c = fgetc(Src_File);
    if (c == 'S')
        type = hex_to_int(fgetc(Src_File));
    else if (c == EOF)
        return 0;
    }
return type;
}

/***********************************************************************/
/* read_s_record: Read the next type 1, 2, or 3 S-record from the      */
/*                input file into the zeroth element in the array of   */
/*                S-record structures.                                 */
/*                                                                     */
/*     RETURNS: OK iff an S-record is found.                           */
/*        NOTE: A fatal error occurs if the S-record contains greater  */
/*              than ARRAY_SIZE data bytes or if the checksum is bad.  */
/***********************************************************************/
static int read_s_record()
{
int type, n;

if (type = find_S123())  /* If there's another s-record.... */
    {
    /*-----------------------------------------------------------------*/
    /* Set the S-record type and initialize the checksum.              */
    /*-----------------------------------------------------------------*/
    S_record[0].type = type;
    Sum = 0;

    /*-----------------------------------------------------------------*/
    /* Get the "length field" of the record, convert it to the number  */
    /* of data bytes (Accounting for the address bytes and the check-  */
    /* sum byte), and check that it is not too big.                    */
    /*-----------------------------------------------------------------*/
    S_record[0].length = nextbyte() - (type + 1) - 1;
    if (S_record[0].length > ARRAY_SIZE)
        {
        printf("Fatal Error: S-record contains > %d data bytes!\n",
               ARRAY_SIZE);
        remove_files();
        exit(-1);
        }

    /*-----------------------------------------------------------------*/
    /* Get the starting address of the record, which could be 2, 3,    */
    /* or 4 character pairs depending on record type.                  */
    /*-----------------------------------------------------------------*/
    S_record[0].address = nextbyte();
    for (n = 0; n < type; n++)
    {
        S_record[0].address = S_record[0].address * 256 + nextbyte();
    }

    /*-----------------------------------------------------------------*/
    /* Fill the structure with the file S-record data.                 */
    /*-----------------------------------------------------------------*/
    for (n = 0; n < S_record[0].length; n++)
        {
        S_record[0].data[n] = nextbyte();
        }

    /*-----------------------------------------------------------------*/
    /* Finally, finish our computation of what the checksum should be  */
    /* and compare it with the value read from the S-record.           */
    /*-----------------------------------------------------------------*/
    Check_Sum = 0xFF & ~Sum;
    if (Check_Sum != nextbyte())
        {
        printf("Fatal Error: Checksum error!\n");
        remove_files();
        exit(-1);
        }
    else
        return OK;
    }

/*---------------------------------------------------------------------*/
/* If there were no more s-records in the file, simply return NOT_OK.  */
/*---------------------------------------------------------------------*/
else
    return NOT_OK;
}

/***********************************************************************/
/* split_to_output_records: Perform the splitting of an S-record read  */
/*                          from the input file into output S-records. */
/*                                                                     */
/* The input S-record is examined one byte at a time.  If a byte falls */
/* into an address range acceptable for one of the target EPROMs, that */
/* byte is copied into that EPROM's S-record structure.  The EPROM     */
/* base address offset of the first byte copied into the S-record      */
/* structure becomes that S-record's address.                          */
/*                                                                     */
/* Command line parameters determine which of the loops below perform  */
/* the actual splitting.  Swapped word and word selections are treated */
/* the same here.  Swapping is handled by the routine that writes the  */
/* S-records to output files.                                          */
/*                                                                     */
/* NOTE: A fatal error occurs here if the user has specified a four-   */
/*       way split and word-wide EPROMs.                               */
/***********************************************************************/
static void split_to_output_records()
{
/*---------------------------------------------------------------------*/
/* Local variable declaration and loop initialization.                 */
/*---------------------------------------------------------------------*/
int n;
int src_index = 0;
int index_1 = 0;
int index_2 = 0;
int index_3 = 0;
int index_4 = 0;
ULONG Address = S_record[0].address;

/*---------------------------------------------------------------------*/
/* All output records are set to the same type as the input record.    */
/*---------------------------------------------------------------------*/
for (n = 1; n < 5; n++)
    S_record[n].type = S_record[0].type;

/*---------------------------------------------------------------------*/
/* One of the following four loops will perform the actual splitting.  */
/*---------------------------------------------------------------------*/
if ((Num_Splits == 2) && (Width == BYTE))
    {
    /*-----------------------------------------------------------------*/
    /* This loop does the splitting for two byte-wide EPROMs.          */
    /*-----------------------------------------------------------------*/
    for (; src_index < S_record[0].length; Address++, src_index++)
        {
        if (even(Address) && in_range(Address))
            {
            if (index_1 == 0)
                S_record[1].address = (Address - Base_Address) >> 1;
            S_record[1].data[index_1++] = S_record[0].data[src_index];
            }
        else if (in_range(Address))  /* must be odd if in_range */
            {
            if (index_2 == 0)
                S_record[2].address = (Address - Base_Address) >> 1;
            S_record[2].data[index_2++] = S_record[0].data[src_index];
            }
        else
            Encountered_out_of_range_data = 1;
        }
    }

else if (Num_Splits == 1)
    {
    /*-----------------------------------------------------------------*/
    /* This loop does the degenerate, one-ROM case.                    */
    /*-----------------------------------------------------------------*/
    for (; src_index < S_record[0].length; Address++, src_index++)
        {
        if (in_range(Address))
            {
            if (index_1 == 0)
                S_record[1].address = (Address - Base_Address);
            S_record[1].data[index_1++] = S_record[0].data[src_index];
            }
        else
            Encountered_out_of_range_data = 1;
        }
    }

else if ((Num_Splits == 4) && (Width == BYTE))
    {
    /*-----------------------------------------------------------------*/
    /* This loop does the splitting for four byte-wide EPROMs.         */
    /*-----------------------------------------------------------------*/
    for (; src_index < S_record[0].length; Address++, src_index++)
        {
        if (((Address & 3) == 0) && in_range(Address))
            {
            if (index_1 == 0)
                S_record[1].address = (Address - Base_Address) >> 2;
            S_record[1].data[index_1++] = S_record[0].data[src_index];
            }
        else if (((Address & 3) == 1) && in_range(Address))
            {
            if (index_2 == 0)
                S_record[2].address = (Address - Base_Address) >> 2;
            S_record[2].data[index_2++] = S_record[0].data[src_index];
            }
        else if (((Address & 3) == 2) && in_range(Address))
            {
            if (index_3 == 0)
                S_record[3].address = (Address - Base_Address) >> 2;
            S_record[3].data[index_3++] = S_record[0].data[src_index];
            }
        else if (in_range(Address))  /* Address & 3 == 3 if in_range */
            {
            if (index_4 == 0)
                S_record[4].address = (Address - Base_Address) >> 2;
            S_record[4].data[index_4++] = S_record[0].data[src_index];
            }
        else
            Encountered_out_of_range_data = 1;
        }
    }
else if ((Num_Splits == 2) && (Width == WORD || Width == WORD_SWAP))
    {
    /*-----------------------------------------------------------------*/
    /* This loop does the splitting for two word-wide EPROMs.          */
    /*-----------------------------------------------------------------*/
    for (; src_index < S_record[0].length; Address++, src_index++)
        {
        if (even(Address >> 1) && in_range(Address))
            {
            if (index_1 == 0)
                S_record[1].address = (Address - Base_Address + 1) >> 1;
            S_record[1].data[index_1++] = S_record[0].data[src_index];
            }
        else if (in_range(Address))  /* must be odd word if in_range */
            {
            if (index_2 == 0)
                S_record[2].address = (Address - Base_Address - 1) >> 1;
            S_record[2].data[index_2++] = S_record[0].data[src_index];
            }
        else
            Encountered_out_of_range_data = 1;
        }
    }
else
    {
    /*-----------------------------------------------------------------*/
    /* A four-way split with word-wise EPROMs has been selected.       */
    /*-----------------------------------------------------------------*/
    printf("A four-way split is not allowed with word-wide EPROMs!\n");
    remove_files();
    exit(-1);
    }

/*---------------------------------------------------------------------*/
/* Set the length field to the value determined in the splitting loop. */
/*---------------------------------------------------------------------*/
S_record[1].length = index_1;
S_record[2].length = index_2;
S_record[3].length = index_3;
S_record[4].length = index_4;
}

/***********************************************************************/
/* update_byte_count: Add length of output S-records to the byte count */
/*                                                                     */
/***********************************************************************/
static void update_byte_count()
{
int n;

for (n = 1; n < 5; n++)
    Byte_Count += S_record[n].length;
}

/***********************************************************************/
/*     putbyte: Write an 8-bit integer value to the Output_File as two */
/*              hex ASCII characters.  Maintain sum of integers        */
/*              written for checksum calculation.                      */
/*                                                                     */
/*      INPUTS: outfile: A pointer to an output file stream.           */
/*              value: The integer to write to the output file.        */
/***********************************************************************/
static void putbyte(value)
int value;
{
fprintf(Output_File, "%02X", value & 0xFF);
Sum += value;
}

/***********************************************************************/
/* put_s_record: Write an S-record structure to the Output_File.       */
/*                                                                     */
/*     NOTE(S):                                                        */
/***********************************************************************/
static void put_s_record(s_record)
S_RECORD *s_record;
{
int i;

/*---------------------------------------------------------------------*/
/* Return immediately if the S-record data length is zero.             */
/*---------------------------------------------------------------------*/
if (s_record->length == 0)
    return;

/*---------------------------------------------------------------------*/
/* Print the first two characters and initialize the checksum.         */
/*---------------------------------------------------------------------*/
fputc('S', Output_File);
fputc(s_record->type + '0', Output_File);
Sum = 0;

/*---------------------------------------------------------------------*/
/* Output the record length field (accounting for the address field    */
/* bytes and the checksum byte).                                       */
/*---------------------------------------------------------------------*/
putbyte(s_record->length + (s_record->type + 1) + 1);

/*---------------------------------------------------------------------*/
/* Send the address field.                                             */
/*---------------------------------------------------------------------*/
switch (s_record->type)
    {
    case 3:  putbyte((s_record->address >> 24) & 0xFF);
    case 2:  putbyte((s_record->address >> 16) & 0xFF);
    default: putbyte((s_record->address >>  8) & 0xFF);
             putbyte(s_record->address & 0xFF);
    }

/*---------------------------------------------------------------------*/
/* Send the data fields.                                               */
/*---------------------------------------------------------------------*/
for (i = 0; i < s_record->length; i++)
    putbyte(s_record->data[i]);

/*---------------------------------------------------------------------*/
/* Send the check-sum and trailing carriage return/line feed.          */
/*---------------------------------------------------------------------*/
putbyte(~Sum);
fputc('\n', Output_File);
}

/***********************************************************************/
/* put_s_record_stub: Write a one byte S-record to the Output_File.    */
/*                                                                     */
/*        NOTE: This routine is required for byte swapping mode.       */
/***********************************************************************/
static void put_s_record_stub(type, byte, stub_address)
int type, byte;
ULONG stub_address;
{
/*---------------------------------------------------------------------*/
/* Print the first two characters and initialize the checksum.         */
/*---------------------------------------------------------------------*/
fputc('S', Output_File);
fputc(type + '0', Output_File);
Sum = 0;

/*---------------------------------------------------------------------*/
/* Output the record length field (accounting for the address field    */
/* bytes and the checksum byte).                                       */
/*---------------------------------------------------------------------*/
putbyte(type + 3);

/*---------------------------------------------------------------------*/
/* Send the address field.                                             */
/*---------------------------------------------------------------------*/
switch (type)
    {
    case 3:  putbyte((stub_address >> 24) & 0xFF);
    case 2:  putbyte((stub_address >> 16) & 0xFF);
    default: putbyte((stub_address >>  8) & 0xFF);
             putbyte(stub_address & 0xFF);
    }

/*---------------------------------------------------------------------*/
/* Send the data, check-sum and trailing carriage return/line feed.    */
/*---------------------------------------------------------------------*/
putbyte(byte);
putbyte(~Sum);
fputc('\n', Output_File);
}

/***********************************************************************/
/*        swap: swap even and odd bytes in an S-record data array.     */
/*                                                                     */
/*      INPUTS: data_ptr: A pointer to some S-record's data array.     */
/*        NOTE: This routine swaps all the data without looking at the */
/*              structure's data length field.                         */
/***********************************************************************/
static void swap(data_ptr)
char *data_ptr;
{
int i, tmp;

for (i = 0; i < ARRAY_SIZE; i += 2, data_ptr += 2)
    {
    tmp = *data_ptr;
    *data_ptr = *(data_ptr + 1);
    *(data_ptr + 1) = tmp;
    }
}

/***********************************************************************/
/* copy_down_1_byte: Move the bytes in an S-record data array down one */
/*                   byte; losing the first byte.                      */
/*                                                                     */
/*      INPUTS: data_ptr: A pointer to some S-record's data array.     */
/*        NOTE: This routine moves all the data without looking at the */
/*              structure's data length field.                         */
/***********************************************************************/
static void copy_down_1_byte(data_ptr)
char *data_ptr;
{
int i;

for (i = 0; i < ARRAY_SIZE - 1; i++, data_ptr++)
    {
    *data_ptr = *(data_ptr + 1);
    }
}

/***********************************************************************/
/* put_swapped_s_record: Write to an output file an S-record structure */
/*                       that is to be byte swapped.                   */
/*                                                                     */
/* There are four cases that must be handled that depend on whether    */
/* the S-record address and the S-record data length are odd or even.  */
/*                                                                     */
/* even address, even length:                0 1 2 3    0 1 2 3        */
/*   Swap bytes and write record.            A B C D => B A D C        */
/*                                                                     */
/* even address, odd length:                 0 1 2 3 4    0 1 2 3   5  */
/*   Swap bytes, dec length, write record,   A B C D E => B A C D & E  */
/*   and write last byte as stub record.                               */
/*                                                                     */
/* odd address, even length:                 1 2 3 4    0   2 3   5    */
/*   Write first byte as stub record, dec    A B C D => A & C B & D    */
/*   length by 2, inc address, copy-down 1                             */
/*   byte, swap, write record, write last byte as stub.                */
/*                                                                     */
/* odd address, odd length:                  1 2 3 4 5    0   2 3 4 5  */
/*   Write first byte as stub record, dec    A B C D E => A & C B E D  */
/*   length, inc address, copy-down 1 byte,                            */
/*   swap bytes, write record.                                         */
/*                                                                     */
/***********************************************************************/
static void put_swapped_s_record(s_record)
S_RECORD *s_record;
{
int last_byte, first_byte;
ULONG stub_address;

/*---------------------------------------------------------------------*/
/* Return immediately if the S-record data length is zero.             */
/*---------------------------------------------------------------------*/
if (s_record->length == 0)
    return;

/*---------------------------------------------------------------------*/
/* Perform one of these four routines to swap and output the S-record. */
/*---------------------------------------------------------------------*/
if (even(s_record->address) && even(s_record->length))
    {
    swap(s_record->data);
    put_s_record(s_record);
    }
else if (even(s_record->address) && odd(s_record->length))
    {
    last_byte = s_record->data[s_record->length - 1];
    stub_address = s_record->address + s_record->length;
    s_record->length--;
    swap(s_record->data);
    put_s_record(s_record);
    put_s_record_stub(s_record->type, last_byte, stub_address);
    }
else if (odd(s_record->address) && even(s_record->length))
    {
    first_byte = s_record->data[0];
    put_s_record_stub(s_record->type, first_byte, s_record->address - 1);
    last_byte = s_record->data[s_record->length - 1];
    stub_address = s_record->address + s_record->length;
    s_record->length -= 2;
    s_record->address++;
    copy_down_1_byte(s_record->data);
    swap(s_record->data);
    put_s_record(s_record);
    put_s_record_stub(s_record->type, last_byte, stub_address);
    }
else /* must be (even(s_record->address) && even(s_record->length)) */
    {
    first_byte = s_record->data[0];
    put_s_record_stub(s_record->type, first_byte, s_record->address - 1);
    s_record->length--;
    s_record->address++;
    copy_down_1_byte(s_record->data);
    swap(s_record->data);
    put_s_record(s_record);
    }
}

/***********************************************************************/
/* write_to_files: Write S-record structures to their output files.    */
/*                                                                     */
/***********************************************************************/
static void write_to_files()
{
int n;

for (n = 1; n <= Num_Splits; n++)
    {
    Output_File = Out_File[n-1];  /* Assign global output file pointer */

    if (Width != WORD_SWAP)
        {
        put_s_record(&S_record[n]);
        }
    else
        {
        put_swapped_s_record(&S_record[n]);
        }
    }
}

/***********************************************************************/
/* close_files: Close all the output files.  Finished writing.         */
/*                                                                     */
/*        NOTE: Before closing the files, we write S9 "termination"    */
/*              records to them.                                       */
/***********************************************************************/
static void close_files()
{
fprintf(Out_File[0], "S9030000FC\n");
fclose(Out_File[0]);
if (Num_Splits > 1)
    {
    fprintf(Out_File[1], "S9030000FC\n");
    fclose(Out_File[1]);
    }

if (Num_Splits == 4)
    {
    fprintf(Out_File[2], "S9030000FC\n");
    fprintf(Out_File[3], "S9030000FC\n");
    fclose(Out_File[2]);
    fclose(Out_File[3]);
    }
}

/***********************************************************************/
/*      report:                                                        */
/*                                                                     */
/*     NOTE(S):                                                        */
/***********************************************************************/
static void report()
{
if (Byte_Count != 0)
    printf("Successfully split input file.  ");
printf("Transferred %ld bytes.\n", Byte_Count);

if (Encountered_out_of_range_data)
    {
    printf("Encountered data that did not map to an EPROM address.\n");
    }
}

/***********************************************************************/
/* M A I N : This program is documented in the comment block at the    */
/*           top of the file.                                          */
/*                                                                     */
/***********************************************************************/
int main(argc, argv)
int argc;
char *argv[];
{
unsigned int rc;

rc = read_params(argc, argv);

if (rc != OK)
    {
    show_usage(rc);
    }
else
    {
    Byte_Count = 0;
    while (read_s_record() == OK)
        {
        split_to_output_records();
        update_byte_count();
        write_to_files();
        }

    close_files();
    report();
    }
exit(0);
}

